#!/usr/bin/env python3
"""
DeRen'Py
by Javantea
Oct 11, 2015

Safely load a Ren'py data.rpa file.
It's a pickle, so we're going to have to safely depickle.
I'm using my hand-written betterPickle which works for renpy files but is not 
yet guaranteed to work with any pickle files.

This program when run from the command line with a rpa file as the argument 
will unpack the file into the current directory.
"""
import zlib
import sys
import io
import os
try:
	import betterPickle
except ImportError:
	print("betterPickle is a necessary dependency for this")
	exit(1)
#end try
loads = betterPickle.loads

def mkdirs(path):
	"""
	Ensure that a directory structure exists.
	FIXME: Deal with a file in the way.
	"""
	dirs = path.split(os.sep)
	dirx = '.' + os.sep
	for d in dirs:
		if not os.path.exists(dirx + d + os.sep):
			os.mkdir(dirx + d)
		#end if
		dirx += d + os.sep
	#next d
#end def mkdirs(path)

def loadRPA(fn, verbose=False):
	"""
	Load an RPA file.
	Code is copied from Ren'Py's actual loader and modified.
	MIT License says that it is okay so here is the copyright just in case.
	# Copyright 2004-2015 Tom Rothamel <pytom@bishoujo.us>
	#
	# Permission is hereby granted, free of charge, to any person
	# obtaining a copy of this software and associated documentation files
	# (the "Software"), to deal in the Software without restriction,
	# including without limitation the rights to use, copy, modify, merge,
	# publish, distribute, sublicense, and/or sell copies of the Software,
	# and to permit persons to whom the Software is furnished to do so,
	# subject to the following conditions:
	#
	# The above copyright notice and this permission notice shall be
	# included in all copies or substantial portions of the Software.
	"""
	prefix = fn[:-4]
	archives = []
	f = open(fn, "rb")
	l = f.readline()

	# 3.0 Branch.
	if l.startswith(b"RPA-3.0 "):
		offset = int(l[8:24], 16)
		key = int(l[25:33], 16)
		#print("xor key:", key)
		f.seek(offset)
		index = loads(zlib.decompress(f.read()), verbose)

		# Deobfuscate the index.
		for k in index.keys():
			if len(index[k][0]) == 2:
				index[k] = [ (offset ^ key, dlen ^ key) for offset, dlen in index[k] ]
			else:
				index[k] = [ (offset ^ key, dlen ^ key, start) for offset, dlen, start in index[k] ]

		archives.append((prefix, index))
		
		f.close()
		return archives

	# 2.0 Branch.
	if l.startswith(b"RPA-2.0 "):
		offset = int(l[8:], 16)
		f.seek(offset)
		index = loads(zlib.decompress(f.read()))
		archives.append((prefix, index))
		f.close()
		return archives

	# 1.0 Branch.
	f.close()
	
	fn = prefix + ".rpi"
	index = loads(zlib.decompress(open(fn, "rb").read()))
	archives.append((prefix, index))
	return archives
#end def loadRPA(fn, verbose=False)

def loadRpyc(fn, verbose=False):
	"""
	Impressive that this can actually reverse a rpyc.
	"""
	data = loads(zlib.decompress(open(fn, "rb").read()), verbose)
	return data
#end def loadRpyc(fn)

def main():
	# Katawa Shoujo is available for free at http://www.katawa-shoujo.com/
	filename = '/home/game/src/Katawa Shoujo-linux-x86/game/data.rpa'
	rpyc = False
	verbose = False
	if '-c' in sys.argv:
		rpyc = True
		sys.argv.remove('-c')
	#end if
	if '-v' in sys.argv:
		verbose = True
		sys.argv.remove('-v')
	#end if
	if '-h' in sys.argv:
		print("Usage: python3 derenpy1.py [-v] [-c filename.rpyc | filename.rpa]")
		print("-v | verbose")
		print("If the file is an rpyc, use the -c flag")
		print("If the file is an rpa, no flag")
		print("If you use the -c flag, it will print out decompilation.")
		print("If you do not, it will load the file as an RPA and create files "
			"and directories \nin the current folder.")
		exit(1)
	#end if
	if len(sys.argv) > 1:
		filename = sys.argv[1]
	#end if
	if rpyc:
		rpy = loadRpyc(filename, verbose)
		print(rpy)
		return 0
	#end if
	archives = loadRPA(filename, verbose)
	if len(archives) > 1:
		print(len(archives), 'archives')
	#end if
	prefix = ''
	index = {}
	for d in archives:
		prefix, index = d
	#end if
	#print(index)
	# Now we have an index. Let's write them all out to here.
	f = open(filename, 'rb')
	for fn in index.keys():
		fnu = fn
		print(fnu)
		if os.path.exists(fnu):
			print("File exists. Will not overwrite.")
			continue
		#end if
		if ('..' + os.sep) in fnu:
			print("Error: Directory traversal attempt:", fnu, file=sys.stderr)
			continue
		#end if
		start = b''
		if len(index[fn][0]) == 2:
			offset, dlen = index[fn][0]
		elif len(index[fn][0]) == 3:
			offset, dlen, start = index[fn][0]
			# Fix the surrogate encoding
			if type(start) == str:
				start = start.encode('utf-8', 'surrogateescape')
			#end if
		else:
			print("Warning: Invalid data:", index[fn], file=sys.stderr)
			continue
		#end if
		f.seek(offset, io.SEEK_SET)
		data = f.read(dlen)
		directory = os.sep.join(fnu.split(os.sep)[:-1])
		if not os.path.exists(directory):
			mkdirs(directory)
		#end if
		open(fnu, 'wb').write(start + data)
	#next fn
#end def main()

if __name__ == '__main__':
	main()
#end if
